/* Copyright (c) 2007 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/** 
 * main.cpp - blue-steel Pong demo
 * @author Brian Sweatt
 * 
 * To be compiled and run on the PPU. This program runs a simple one-player version of pong.
 * Most of the code is for interfacing with the joystick (six-axis) and keyboard, as well as 
 * the simulation of the ball (very simple bouncing, no real physics).
 * Meant to show the ease of use of the blue-steel API.
 * (the rendering code takes up little more than 10 lines)
 * Enjoy!
 */
extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <libvector.h>
#include <simdmath.h>
#include <libspe.h>
#include <libmisc.h>
#include <unpack_rgba8_v.h>
#include <math.h>
#include </usr/include/linux/joystick.h>
}

#include "../src/sphere.h"
#include "../src/common.h"
#include "../src/rgbpacket.h"
#include "../src/fb.h"
#include "../src/objectset.h"
#include "bsapi.h"


//////////////////////////
// Code from keyboard.h //
//////////////////////////

#define INITKEYBOARD ttycbreak()
#define DEINITKEYBOARD ttynorm()

/*   Keyboard support functions. January 7 1993 */

/* obligatory includes. */
#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>                                  
#ifndef FIONREAD
#define FIONREAD TIOCINQ
#endif
#include <signal.h>

static struct termios itio, tio;

static int ttyfd  = -1;
static int gottio = 0;

/* Catch quit signals so we dont end up with a CBREAK terminal when
   someone does a ^c during getch() */
void sig1 (int sig)
{
    tcsetattr (ttyfd, TCSANOW, &itio);
    signal (sig, SIG_DFL);
    kill (getpid(), sig);
}

/* initialize signals */
void setsigs()
{  
    signal (SIGINT,  sig1);
    signal (SIGTSTP, sig1);
    signal (SIGQUIT, sig1);
}

/* remove signals */
void rmsigs()
{
    signal (SIGCONT, SIG_DFL);
    signal (SIGINT,  SIG_DFL);
    signal (SIGTSTP, SIG_DFL);
    signal (SIGQUIT, SIG_DFL);
}

/* Put terminal in CBREAK mode */
int ttycbreak()
{
    if (ttyfd == -1)
	if ((ttyfd = open("/dev/tty", O_RDWR)) < 0)
	    return 0;
    if (tcgetattr(ttyfd, &tio) < 0)
	return 0;
    itio = tio;
    setsigs();
    gottio = 1;
    tio.c_cc[VMIN] = 0;
    tio.c_cc[VTIME] = 0;
    tio.c_lflag &= ~(ECHO|ICANON);
    tcsetattr (ttyfd, TCSANOW, &tio);
    return 1;
}

/* Restore terminal */
int ttynorm()
{
    gottio = 0;
    tcsetattr (ttyfd, TCSANOW, &itio);
    rmsigs();
    return 1;
}

/* Unix kbhit() */
int kbhit()
{
    long lf;
    if (ioctl(ttyfd, FIONREAD, &lf) == -1)
	return 0;
    return lf;
}

/* Unix getch() */
unsigned char getch()
{
    char c;
    read(ttyfd, &c, 1);
    return (unsigned char) c;
}

/////////////////////////
// End keyboard.h code //
/////////////////////////


ObjectSet objects;

vector unsigned int image[IMG_WIDTH*IMG_HEIGHT / 4] __attribute__((aligned(128)));

int main(int argc, char** argv) {
  dprintf("blue-steel parallel raytracer v0.1\r\n");
  if(argc > 1) {
    printf("%s", argv[1]);
    int spes = atoi(argv[1]);
    if(spes > 0 && spes <= TOTAL_SPES) {
        NUM_RENDERING_SPES = spes;
    }
  }
  INITKEYBOARD;
  
  dprintf("Created spe threads\n");
  uint32_t scene_size = (IMG_WIDTH * IMG_HEIGHT * sizeof(vector unsigned int));

  ObjectSet *objects = (ObjectSet *) malloc_align(sizeof(ObjectSet), 7);
  
  // One sphere for the ball... Not that we are setting the material to be refractive and bumpy
  objects->num_spheres = 1;
  objects->spheres = (Sphere *) malloc_align(sizeof(Sphere)*objects->num_spheres, 7);
  objects->spheres[0] = Sphere((vector float) {0.5f, 0.5f, 0.041f, 0.04f}, 5);
  
  // We're using 4 triangles... two for the ground, and two for the paddle
  objects->num_triangles = 4;
  objects->triangles = (Triangle *) malloc_align(sizeof(Triangle)*objects->num_spheres, 7);
  objects->triangles[0] = Triangle((vector float) {0.0f, 0.0f, 0.0f, 0.0f}, 
				   (vector float) {1.0f, 0.0f, 0.0f, 0.0f}, 
				   (vector float) {1.0f, 1.0f, 0.0f, 0.0f},
				   4);
  objects->triangles[1] = Triangle((vector float) {0.0f, 0.0f, 0.0f, 0.0f},
				   (vector float) {1.0f, 1.0f, 0.0f, 0.0f},
				   (vector float) {0.0f, 1.0f, 0.0f, 0.0f},
				   4);
  objects->triangles[2] = Triangle((vector float) {0.4f, 0.02f, 0.01f, 0.0f}, 
				   (vector float) {0.55f, 0.02f, 0.08f, 0.0f},
				   (vector float) {0.4f, 0.02f, 0.08f, 0.0f},
				   2);
  objects->triangles[3] = Triangle((vector float) {0.4f, 0.02f, 0.01f, 0.0f},
				   (vector float) {0.55f, 0.02f, 0.01f, 0.0f}, 
				   (vector float) {0.55f, 0.02f, 0.08f, 0.0f}, 
				   2);

  // Defintions of the materials being used....
  // Note that not all of these are used in this program, but are left as an example of how to 
  // create different types of materials
  objects->num_materials = 9;
  objects->materials = (material *) malloc_align(sizeof(material)*objects->num_materials, 7);
  
  objects->materials[0].materialType = MATERIAL_PHONG;
  objects->materials[0].diffuse_r = 0.0f;
  objects->materials[0].diffuse_g = 0.0f;
  objects->materials[0].diffuse_b = 0.0f;
  objects->materials[0].arg1_i = 0;
  objects->materials[0].arg1_f = 0.0f;
  objects->materials[0].arg2_f = 0.0f;
  objects->materials[0].specular_r = 0.0f;
  objects->materials[0].specular_g = 0.0f;
  objects->materials[0].specular_b = 0.0f;
  objects->materials[0].reflective_r = 0.0f;
  objects->materials[0].reflective_g = 0.0f;
  objects->materials[0].reflective_b = 0.0f;
  objects->materials[0].transparent_r = 0.0;
  objects->materials[0].transparent_g = 0.0;
  objects->materials[0].transparent_b = 0.0;
  
  objects->materials[2].materialType = MATERIAL_PHONG;
  objects->materials[2].diffuse_r = 0.0f;
  objects->materials[2].diffuse_g = 0.5f;
  objects->materials[2].diffuse_b = 0.7f;
  objects->materials[2].arg1_i = 64;
  objects->materials[2].specular_r = 1.0f;
  objects->materials[2].specular_g = 1.0f;
  objects->materials[2].specular_b = 1.0f;
  objects->materials[2].reflective_r = 0.0f;
  objects->materials[2].reflective_g = 0.5f;
  objects->materials[2].reflective_b = 0.8f;
  objects->materials[2].transparent_r = 0.0f;
  objects->materials[2].transparent_g = 0.0f;
  objects->materials[2].transparent_b = 0.0f;
  objects->materials[3].materialType = MATERIAL_PHONG;
  objects->materials[3].diffuse_r = 0.1f;
  objects->materials[3].diffuse_g = 0.1f;
  objects->materials[3].diffuse_b = 0.6f;
  objects->materials[3].arg1_i = 128;
  objects->materials[3].arg1_f = 1.05f;
  objects->materials[3].specular_r = 1.0f;
  objects->materials[3].specular_g = 1.0f;
  objects->materials[3].specular_b = 1.0f;
  objects->materials[3].reflective_r = 0.4f;
  objects->materials[3].reflective_g = 0.4f;
  objects->materials[3].reflective_b = 0.4f;
  objects->materials[3].transparent_r = 0.9f;
  objects->materials[3].transparent_g = 0.9f;
  objects->materials[3].transparent_b = 0.9f;
  objects->materials[4].materialType = MATERIAL_BUMP;
  objects->materials[4].arg1_i = 1;
  objects->materials[4].arg1_f = 0.4f;
  objects->materials[4].arg2_f = 0.2f;
  objects->materials[5].materialType = MATERIAL_BUMP;
  objects->materials[5].arg1_i = 3;
  objects->materials[5].arg1_f = 0.8f;
  objects->materials[5].arg2_f = 0.3f;

  objects->materials[1].materialType = MATERIAL_CHECKERBOARD;
  objects->materials[1].colorDSRT_r = (vector float){10.f, 0.0f, 0.0f, 0.0f};
  objects->materials[1].colorDSRT_g = (vector float){0.0f, 10.0f, 0.0f, 0.0f};
  objects->materials[1].colorDSRT_b = (vector float){0.0f, 0.0f, 10.0f, 0.0f};
  objects->materials[1].colorDSRT_a = (vector float){0.0f, 0.0f, 0.0f, 10.0f};
  objects->materials[1].arg1_i = 7;
  objects->materials[1].arg2_i = 8;
  
 
  objects->materials[6].materialType = MATERIAL_WOOD;
  objects->materials[6].colorDSRT_r = (vector float){10.0f, 0.0f, 0.0f, 0.0f};
  objects->materials[6].colorDSRT_g = (vector float){0.0f, 100.0f, 0.0f, 0.0f};
  objects->materials[6].colorDSRT_b = (vector float){0.0f, 0.0f, 100.0f, 0.0f};
  objects->materials[6].colorDSRT_a = (vector float){0.0f, 0.0f, 0.0f, 50.0f};
  objects->materials[6].arg1_f = 1.0f; //frequency
  objects->materials[6].arg2_f = 5.0f; //amplitude
  objects->materials[6].arg1_i = 7;
  objects->materials[6].arg2_i = 8;
  
  objects->materials[7].materialType = MATERIAL_PHONG;
  objects->materials[7].diffuse_r = 0.0f;
  objects->materials[7].diffuse_g = 0.0f;
  objects->materials[7].diffuse_b = 0.0f;
  objects->materials[7].arg1_i = 128;
  objects->materials[7].arg1_f = 2.1f;
  objects->materials[7].specular_r = 1.0f;
  objects->materials[7].specular_g = 1.0f;
  objects->materials[7].specular_b = 1.0f;
  objects->materials[7].reflective_r = 0.8f;
  objects->materials[7].reflective_g = 0.8f;
  objects->materials[7].reflective_b = 0.8f;
  objects->materials[7].transparent_r = 0.0f;
  objects->materials[7].transparent_g = 0.0f;
  objects->materials[7].transparent_b = 0.0f;

  objects->materials[8].materialType = MATERIAL_PHONG;
  objects->materials[8].diffuse_r = 0.8f;
  objects->materials[8].diffuse_g = 0.8f;
  objects->materials[8].diffuse_b = 0.8f;
  objects->materials[8].arg1_i = 128;
  objects->materials[8].arg1_f = 2.1f;
  objects->materials[8].specular_r = 0.5f;
  objects->materials[8].specular_g = 0.5f;
  objects->materials[8].specular_b = 0.5f;
  objects->materials[8].reflective_r = 0.0f;
  objects->materials[8].reflective_g = 0.0f;
  objects->materials[8].reflective_b = 0.0f;
  objects->materials[8].transparent_r = 0.0f;
  objects->materials[8].transparent_g = 0.0f;
  objects->materials[8].transparent_b = 0.0f;
  
  objects->num_lights = 1;
  objects->lights = (light *) malloc_align(sizeof(light)*objects->num_lights, 7);
  // Commented out, but gives an alternate implementation using a point light source...
  // Later in the code, we update the position of the light to follow the paddle.
  // This has no effect for the directional light source, but will show with the point light
  /*
  objects->lights[0].type = LIGHT_POINT;
  
  objects->lights[0].position_x = 0.5f;
  objects->lights[0].position_y = 1.0f; 
  objects->lights[0].position_z = 0.1f;
  
  objects->lights[0].color_x = 1.0f;
  objects->lights[0].color_y = 0.0f;
  objects->lights[0].color_z = 0.0f;
  objects->lights[0].attn_x = 1.0f;
  objects->lights[0].attn_y = 1.0f;
  objects->lights[0].attn_z = 1.0f;
  */
  objects->lights[0].type = LIGHT_DIRECTIONAL;
  objects->lights[0].direction_x = 0.0f;
  objects->lights[0].direction_y = 1.0f;
  objects->lights[0].direction_z = -1.0f;
  objects->lights[0].color_x = 1.0f;
  objects->lights[0].color_y = 1.0f;
  objects->lights[0].color_z = 1.0f;
  objects->lights[0].attn_x = 1.0f;
  objects->lights[0].attn_y = 1.0f;
  objects->lights[0].attn_z = 1.0f;
  
  bsInit(objects);
  

  //INITIALIZE FRAME BUFFER
  
  bool fb_good = false;
  if(fb_good = fb_init())
    fb_set_all_pixels(RGB32(0, 0, 0));
  
  // INITIALIZE JOYSTICK
  
  int fd = open("/dev/js0", O_RDONLY | O_NONBLOCK);

  float cam_x = 1.0f;
  float cam_y = -0.5;
  float cam_z = 1.5f;
  int direction = 1;
  
  float lightDir = 0.1f;
  float bx, by, bz;
  bx = 0.3f; by = 0.8f;  bz = 0.041f;
  float dx,dy,dz;
  dx = 0.006f; dy = 0.014f; dz = 0.0f;
  

  // Paddle will move in the x direction
  float paddleSpeed[2] = {0.02f, -0.02f};
  float paddleX = 0.4f;

  int inum = 1;
  float totalTime = 0;
  bool paused = false;
  // 4 view modes: top of board looking at ball, paddle looking at ball, ball looking at paddle, above board
  int viewMode = 3;
  
  // Set the initial position of the sphere, as well as the initial camera position and direction
  // and the ambient lighting term
  objects->spheres[0].setPosition((vector float) {bx, by, bz, 0.0f}); 
  
  bsSetCenter(0.5f, 1.2f, 0.2f);
  bsLookAt(bx, by, bz, 0.0f, 0.0f, 1.0f);
  bsAmbient(0.1f, 0.1f, 0.1f);

  while(1) {
    
    // If we're paused, we still want to receive input to unpause, but nothing else...
    // So we check for input, but simply continue with the next iteration of the game loop
    // if we haven't been unpaused. This check it done within a frame rendering, but 
    // the latency from the rendering process is not a problem in practice, where the user
    // will receive almost immediate unpausing, despite waiting on the scene to finish rendering
    if (paused) {
      bsRenderScene(image);
      if (getch() == 'p') 
	paused = false;

      struct js_event pEvent;

      int s = read(fd, &pEvent, sizeof(struct js_event));
      
      if (s > 0) {
	int type = pEvent.type;
	type &= ~JS_EVENT_INIT;
	
	if (type == JS_EVENT_BUTTON && pEvent.number == 3 && pEvent.value)
	  paused = false;
      }
      bsWait();
      
      // Still want to update the frame buffer when paused, this assures that 
      if(fb_good) {
	fb_memcpy_frame((unsigned int*)&image[0], IMG_WIDTH, IMG_HEIGHT);
	fb_flip_buffer();
      }

      continue;
    }
    uint32_t curr_i = 0;

    // Let's get things started...
    // Start the scene rendering process on the SPUs
    bsRenderScene(image);

    // Perform the updates to the ball velocities if we have hit the extremities
    bx += dx;
    by += dy;
    bz += dz;   
    
    if ((bx > paddleX) && (bx < paddleX + 0.15f) && (by < 0.09f)) {
      dy = -dy;
      by = 0.09f;
    }

    if ((bx > 0.96f) || (bx < 0.04f)){
      dx = -dx;
    }
    
    if ((by > 0.96f)) {
      dy = -dy;
    }

    if (by < 0.04f) {
      paused= true;
      dy = -dy;
      bx = 0.3f;
      by = 0.8f;
    }
    
    char input = '0';
    
    struct js_event jsEvent;
    static bool lDown = false;
    static bool rDown = false;

    // While the scene is still being rendered... Accept keyboard and joystick input
    while (!bsDone()) {
      if (kbhit()) {
	input = getch();
      }
      
      int s = read(fd, &jsEvent, sizeof(struct js_event));
      if (s > 0) {
	int type = jsEvent.type;
	type &= ~JS_EVENT_INIT;
      
	if (type == JS_EVENT_BUTTON) {
	  printf("Button pressed: %i\n", jsEvent.number);

	  if (jsEvent.number == 5) {
	    rDown = (bool) jsEvent.value;
	  }
	  if (jsEvent.number == 7) {
	    lDown = (bool) jsEvent.value;
	  }
	  if (jsEvent.number == 10 && jsEvent.value) {
	    viewMode = (viewMode > 0)? (viewMode - 1):0;
	  }
	  if (jsEvent.number == 11 && jsEvent.value) {
	    viewMode = (viewMode < 3)? (viewMode + 1):3;
	  }
	  if (jsEvent.number == 16 && jsEvent.value) {
	    DEINITKEYBOARD;
	    exit(0);
	  }
	  if (jsEvent.number == 3 && jsEvent.value) {
	    paused = true;
	  }
	}
      }
    }

    // Wait on the rendering to finish
    bsWait();

    // After the rendering has completed, let's update the objects based on the input received
    int whichSpeed = viewMode%2;
    bool checkBounds[2] = {(paddleX > 0.0f), (paddleX < 1.0f - 0.2f)};

    if (rDown && checkBounds[whichSpeed]) {
      paddleX -= paddleSpeed[whichSpeed];
      objects->triangles[2].translate((vector float) {-paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
      objects->triangles[3].translate((vector float) {-paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
    }

    if (lDown && checkBounds[1 - whichSpeed]) {
      paddleX += paddleSpeed[whichSpeed];
      objects->triangles[2].translate((vector float) {paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
      objects->triangles[3].translate((vector float) {paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
    }
      

    switch (input) {
    case '/':
      if (checkBounds[whichSpeed]) {
	paddleX -= -paddleSpeed[whichSpeed];
	objects->triangles[2].translate((vector float) {-paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
	objects->triangles[3].translate((vector float) {-paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
      }
      break;
    case 'z':
      if (checkBounds[whichSpeed]) {
	paddleX += paddleSpeed[whichSpeed];
	objects->triangles[2].translate((vector float) {paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
	objects->triangles[3].translate((vector float) {paddleSpeed[whichSpeed], 0.0f, 0.0f, 0.0f});
      }
      break;
    case 'p':
      paused = true;
      break;
    case '1':
      viewMode = 0;
      break;
    case '2':
      viewMode = 1;
      break;
    case '3':
      viewMode = 2;
      break;
    default:
      break;
    }
    
    objects->spheres[0].setPosition((vector float) {bx, by, bz, 0.0f}); 
    objects->lights[0].position_x = paddleX + 0.1f;
  
    if (viewMode == 0) {
      bsSetCenter(0.5f, 1.2f, 0.2f);
      bsLookAt(bx, by, bz, 0.0f, 0.0f, 1.0f);
    }
      

    if (viewMode == 1) {
      bsSetCenter(paddleX + 0.05f, 0.0f, 0.15f);
      bsLookAt(bx, by, bz, 0.0f, 0.0f, 1.0f);
    }

    if (viewMode == 2) {
      bsSetCenter(bx, by+0.15f, bz + 0.1f);
      bsLookAt(bx, 0.05f, 0.0f, 0.0f, 0.0f, 1.0f);
    }

    if (viewMode == 3) {
      bsSetCenter(0.5f, 0.5f, 1.0f);
      bsLookAt(0.5f, 0.5f, 0.0f, 0.0f, 1.0f, 0.0f);
    }

    // If the frame buffer is not being used, we can write each frame to an image file...
    // This code is commented out in the release, but is left in to demonstrate the process 
    // an image would undergo for writing the pixels.
    /*
#if(!FRAME_BUFFER)        
    int index = 0;
    for (int i = IMG_HEIGHT - 1; i >= 0 ; i--) {
      for (int j = 0; j < IMG_WIDTH; j+=4) {
	
	//union {
	//vector float c;
	//float v[4];
	//} rv, gv, bv;
	//rv.c = image[index].r; gv.c = image[index].g; bv.c = image[index].b;
	//dprintf("(i,j) (r,g,b):\t(%i, %i)\t (%f, %f, %f)", i, j, rv.v[0], gv.v[0], bv.v[0]);
	
	
	uint32_t x = j;
	uint32_t y = i;
	
	vector unsigned int xx = (vector unsigned int) {x, x + 1, x + 2, x + 3};
	vector unsigned int yy = (vector unsigned int) {y, y, y, y};
	RGBPacket color;
	color.r = _unpack_rgba8_v(image[index], 0);
	color.g = _unpack_rgba8_v(image[index], 1);
	color.b = _unpack_rgba8_v(image[index], 2);
	img->SetPixelPacket(xx, yy, color);
	
	index++;
      }
    }
    
    sprintf(filename, "demo%i.tga", inum);
    
    img->SaveTGA(filename);
    
#endif
    */
    inum++;

    if(fb_good) {
      fb_memcpy_frame((unsigned int*)&image[0], IMG_WIDTH, IMG_HEIGHT);
      fb_flip_buffer();
    }
  } //render loop
  
  DEINITKEYBOARD;
  
  if(fb_good) {
    fb_shutdown();
    fb_good = false;
  }
  
  return 0;
}
