/* Copyright (c) 2007 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/**
 * raytracer.cpp - Implementation for routines of the RayTracer class
 * @author Brian Sweatt
 *
 * The CLIPPING_T, LIGHT_CLIPPING, and CLIPPING_WEIGHT parameters can be tweaked in this file...
 * These are the maximum intersection time that will recurse for reflection and refraction, the 
 * minimum lighting contribution before a shadow is assumed, and the minimum recursion weight, respectively.
 * 
 * Altering these parameters by decreasing CLIPPING_T, increasing LIGHT_CLIPPING, or increasing CLIPPING_WEIGHT
 * may increase performance, but will decrease rendering quality. Alter at your own rendering risk.
 */

extern "C" {
#include <reflect_vec3_v.h>
#include <dot_product3_v.h>
#include <dot_product3.h>
#include <normalize3_v.h>
#include <spu_mfcio.h>
  // #include <simdmath.h>
#include <simdmath/recipf4.h>
#include <simdmath/sqrtf4.h>
}
#include "raytracer.h"
#include "light.h"
#include "material.h"
#include "common.h"


#define CLIPPING_T 1000.0f
// Warning: setting light clipping could lead to ugly artifacts if set too high....
#define LIGHT_CLIPPING 0.0f
#define CLIPPING_WEIGHT 0.0f

RayTracer :: RayTracer(ObjectSet *Objects, int Bounces, bool Shadows, int NumLights, vector float Amb): 
  objects(Objects), 
  bounces(Bounces), 
  shadows(Shadows), 
  num_lights(NumLights),
  ambient_light(Amb) 
{
}

int RayTracer :: getBounces() const {
  return bounces;
}

// Returns the mask for which directions are valid.... Invalid directions are filled in with ZERO
vector unsigned int RayTracer::transmit_vec3_v(vector float &dx, vector float &dy, vector float &dz,
					       const vector float nx, const vector float ny, const vector float nz,
					       const vector float ix, const vector float iy, const vector float iz, 
					       vector float index_i, vector float index_t) const {
  vector float Ix, Iy, Iz;
  vector float nr = _recipf4_fast(index_t);
  const vector unsigned int sign_bit = spu_splats((unsigned int)0x80000000);
  Ix = (vector float) spu_xor((vector unsigned int) ix, sign_bit);
  Iy = (vector float) spu_xor((vector unsigned int) iy, sign_bit);
  Iz = (vector float) spu_xor((vector unsigned int) iz, sign_bit);
  nr = spu_mul(nr, index_i);
  vector float nDotI = _dot_product3_v(nx, ny, nz, Ix, Iy, Iz);
  vector float one_v = spu_splats(1.0f);
  vector float sqrtTerm = spu_nmsub(nDotI, nDotI, one_v);
  sqrtTerm = spu_nmsub(spu_mul(nr, nr), sqrtTerm, one_v);
  vector float zero_v = spu_splats(0.0f);
  vector unsigned int valid = spu_cmpgt(sqrtTerm, zero_v);
  vector float rx, ry, rz;
  vector float rfactor;
  sqrtTerm = _sqrtf4(sqrtTerm);
  rx = spu_mul(nr, Ix);
  ry = spu_mul(nr, Iy);
  rz = spu_mul(nr, Iz);
  rfactor = spu_msub(nr, nDotI, sqrtTerm);
  rx = spu_msub(rfactor, nx, rx);
  ry = spu_msub(rfactor, ny, ry);
  rz = spu_msub(rfactor, nz, rz);
  _normalize3_v(&rx, &ry, &rz, rx, ry, rz);
  dx = spu_sel(zero_v, rx, valid);
  dy = spu_sel(zero_v, ry, valid);
  dz = spu_sel(zero_v, rz, valid);
  
  return valid;
}

RGBPacket RayTracer :: tracePacket(const RayPacket r, HitPacket &hit, float tmin, int bounces, vector float refIndex) const {
  RGBPacket color;
  vector float old_t = hit.t;
  vector float zero_v = spu_splats(0.0f);
  vector float one_v = spu_splats(1.0f);
  vector float half = spu_splats(0.5f);
  vector float max_v = spu_splats(MAX_FLOAT);
  /*  vector float rx0 = r.x0;
  vector float ry0 = r.y0;
  vector float rz0 = r.z0;
  vector float rdx = r.dx;
  vector float rdy = r.dy;
  vector float rdz = r.dz;
  vector float hit_t = hit.t;
  vector float hit_nx = hit.nx;
  vector float hit_ny = hit.ny;
  vector float hit_nz = hit.nz;
  */


  color.r = color.g = color.b = zero_v;
  
  //printf("TracePacket: %i %i  %i\n", objects->num_spheres, objects->num_triangles, bounces);


  bool sAlloc = (bool) objects->spheres;
  bool tAlloc = (bool) objects->triangles;

  // DMA transfers have been started on the 
  // Loop over all the objects in the scene, one primitive-type at a time
  mfc_write_tag_mask(1 << DMA_TAG_TRIANGLES);
  mfc_read_tag_status_all();
  //printf("TracePacket... finished DMAing triangles\n");
  for (int j = 0; tAlloc && (j < objects->num_triangles); j++) {
    (*(objects->triangles+j)).intersectPacket(r, hit, tmin);
  }
  mfc_write_tag_mask(1 << DMA_TAG_SPHERES);
  mfc_read_tag_status_all();
  for (int j = 0; sAlloc && (j < objects->num_spheres); j++) {
    (*(objects->spheres+j)).intersectPacket(r, hit, tmin);
  }

  vector unsigned int changed = spu_cmpgt(old_t, hit.t);
  vector float x = spu_madd(r.dx, hit.t, r.x0);
  vector float y = spu_madd(r.dy, hit.t, r.y0);
  vector float z = spu_madd(r.dz, hit.t, r.z0);
  
  mfc_write_tag_mask(1 << DMA_TAG_LIGHTS);
  mfc_read_tag_status_all();
  // After finding the closest intersection, shade it for every light in the scene
  for (int i = 0; i < objects->num_lights; i++) {
    vector float dirToLightx, dirToLighty, dirToLightz, distanceToLight;
    RGBPacket light_color;
    RGBPacket curr;
    light_color.r = zero_v; light_color.g = zero_v; light_color.b = zero_v;
    curr.r = zero_v; curr.g = zero_v; curr.b = zero_v;
    RayPacket sray;
    HitPacket shit;
    sray.x0 = x; sray.y0 = y; sray.z0 = z;    

    light_getIllumination(*(objects->lights+i), light_color, dirToLightx, dirToLighty, dirToLightz, 
			  distanceToLight, x, y, z);
#if(SHADING)
#if(SHADOWS)
    // The following is code to see if the contribution from a light is significant enough to make 
    // a noticeable impact on the scene... This is adjusted using the LIGHT_CLIPPING constant
    // The branch resulting from the if statement isn't costly in most cases, where this will 
    // cut out alot of work intersecting primitives to check for shadows
    
    vector float lightclip = spu_splats((float)LIGHT_CLIPPING);
    vector unsigned int magltclip = spu_cmpgt(lightclip, light_color.r);
    magltclip = spu_and(magltclip, spu_cmpgt(lightclip, light_color.g));
    magltclip = spu_and(magltclip, spu_cmpgt(lightclip, light_color.b));
    magltclip = spu_gather(magltclip);
    
    if (spu_extract(magltclip, 0) == 15)
      continue;
    shit.t = distanceToLight;
    shit.nx = shit.ny = shit.nz = zero_v;
    // Assuming dirToLight is normalized...
    sray.dx = dirToLightx; sray.dy = dirToLighty; sray.dz = dirToLightz;
#endif //SHADOWS
    mfc_write_tag_mask(1 << DMA_TAG_MATERIALS);
    mfc_read_tag_status_all();
    
    materialShade(objects->materials, curr, r, hit, 
		  dirToLightx, dirToLighty, dirToLightz, 
		  light_color.r, light_color.g, light_color.b,
		  changed);
#if(SHADOWS)    
    // Check all primitives for light occlusion
    for (int j = 0; sAlloc && (j < objects->num_spheres); j++) {
      (*(objects->spheres+j)).intersectPacket(sray, shit, 0.01f);
    }
    vector unsigned int occluded = spu_cmpgt(distanceToLight, shit.t);
    int count  = spu_extract(spu_gather(occluded), 0);
    //for (int j = 0; (count < 15) && (objects->cylinders) && (j < objects->num_cylinders); j++) {
    //  objects->cylinders[j].intersectPacket(sray, shit, 0.0001f);
    //}
    for (int j = 0; (count < 15) && tAlloc && (j < objects->num_triangles); j++) {
      (*(objects->triangles+j)).intersectPacket(sray, shit, 0.01f);
    }
    occluded = spu_cmpgt(distanceToLight, shit.t);
    
    color.r = spu_sel(spu_add(color.r, curr.r), color.r, occluded);
    color.g = spu_sel(spu_add(color.g, curr.g), color.g, occluded);
    color.b = spu_sel(spu_add(color.b, curr.b), color.b, occluded);
#else
    color.r = spu_add(color.r, curr.r);
    color.g = spu_add(color.g, curr.g);
    color.b = spu_add(color.b, curr.b);
#endif //SHADOWS
    
#else
    vector float grayShade = _dot_product3_v(hit.nx, hit.ny, hit.nz, dirToLightx, dirToLighty, dirToLightz);
    // TODO: FINISH THIS, YOU LEFT FOR DINNER!!!!
    color.r = color.g = color.b = grayShade;
#endif
    
  }
  
  // add the ambient lighting term

  vector unsigned char splat0 = (vector unsigned char) {0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3};
  vector unsigned char splat1 = (vector unsigned char) {4,5,6,7,4,5,6,7,4,5,6,7,4,5,6,7};
  vector unsigned char splat2 = (vector unsigned char) {8,9,10,11,8,9,10,11,8,9,10,11,8,9,10,11};
 
  // Compute and add the ambient term to the color, regardless of light sources
#if(SHADING)
  RGBPacket amb_term;
  RGBPacket diffuse_color;
  diffuse_color.r = diffuse_color.g = diffuse_color.b = zero_v;
  material_getDiffuseColor(objects->materials, diffuse_color, hit, x, y, z, changed);

  amb_term.r = spu_shuffle(ambient_light, ambient_light, splat0);
  amb_term.g = spu_shuffle(ambient_light, ambient_light, splat1);
  amb_term.b = spu_shuffle(ambient_light, ambient_light, splat2);
  amb_term.r = spu_mul(amb_term.r, diffuse_color.r);
  amb_term.g = spu_mul(amb_term.g, diffuse_color.g);
  amb_term.b = spu_mul(amb_term.b, diffuse_color.b);
  amb_term.r = spu_add(color.r, amb_term.r);
  amb_term.g = spu_add(color.g, amb_term.g);
  amb_term.b = spu_add(color.b, amb_term.b);  

  color.r = spu_sel(color.r, amb_term.r, changed);
  color.g = spu_sel(color.g, amb_term.g, changed);
  color.b = spu_sel(color.b, amb_term.b, changed);
#endif

  
  // Only add this code to the method if we're bouncing, since it has an inefficient, hard-to-predict branch
#if (MAX_BOUNCES>0)
  
  vector float clipweight = spu_splats((float)CLIPPING_WEIGHT);  

  
  if ((bounces > 0)) {
  
    vector unsigned int sign_bit = spu_splats((unsigned int) 0x80000000);
    vector float rayDotn = _dot_product3_v(hit.nx, hit.ny, hit.nz, r.dx, r.dy, r.dz);
    RGBPacket reflectiveColor;
    reflectiveColor.r = reflectiveColor.g = reflectiveColor.b = zero_v;
    material_getReflectiveColor(objects->materials, reflectiveColor, hit, x, y, z, changed);
    material_bumpNormal(objects->materials, hit, x,y,z,changed);
    vector unsigned int colorgtclip = spu_cmpgt(reflectiveColor.r, clipweight);
    colorgtclip = spu_or(colorgtclip, spu_cmpgt(reflectiveColor.g, clipweight));
    colorgtclip = spu_or(colorgtclip, spu_cmpgt(reflectiveColor.b, clipweight));
    colorgtclip = spu_gather(colorgtclip);

    vector float clipt_v = spu_splats((float)CLIPPING_T);
    vector unsigned int reflectiver = spu_cmpgt(reflectiveColor.r, zero_v);
    vector unsigned int reflectiveg = spu_cmpgt(reflectiveColor.g, zero_v);
    vector unsigned int reflectiveb = spu_cmpgt(reflectiveColor.b, zero_v);
    vector unsigned int  reflective = spu_or(reflectiver, reflectiveg);
    vector unsigned int frontSide = spu_cmpgt(zero_v, rayDotn);	
    vector unsigned int clipgthit = spu_cmpgt(clipt_v, hit.t);
    reflective = spu_or(reflective, reflectiveb);
    // Make sure we actually have a hit at this point...
    reflective = spu_and(reflective, changed);
    //clipgthit = spu_gather(clipgthit);
    //reflective = spu_and(reflective, spu_cmpgt(spu_splats((float)bounces), zero_v));
    // As we trace packets, we must trace one whole packet even if only one ray reflects... bummer.
    reflective = spu_and(reflective, clipgthit);
    //reflective = spu_and(reflective, colorgtclip);
    reflectiveb = spu_gather(reflective);
    int recurse = (spu_extract(reflectiveb, 0) > 0);
    // recurse for reflection...
    if (recurse) {
      RayPacket ref;
      HitPacket refHit;
      RGBPacket bounceColor;
      vector float rnx, rny, rnz;
      rnx = (vector float) spu_xor((vector unsigned int)hit.nx, sign_bit);
      rny = (vector float) spu_xor((vector unsigned int)hit.ny, sign_bit);
      rnz = (vector float) spu_xor((vector unsigned int)hit.nz, sign_bit);
      rnx = spu_sel(rnx, hit.nx, frontSide);
      rny = spu_sel(rny, hit.ny, frontSide);
      rnz = spu_sel(rnz, hit.nz, frontSide);
      _reflect_vec3_v(&ref.dx, &ref.dy, &ref.dz,r.dx,r.dy, r.dz, rnx, rny, rnz);
      ref.dx = spu_sel(zero_v, ref.dx, reflective);
      ref.dy = spu_sel(zero_v, ref.dy, reflective);
      ref.dz = spu_sel(zero_v, ref.dz, reflective);
      // Shouldn't have to spu_sel the x,y, and z, since a ray with zero direction shouldn't hit anything anyhow
      ref.x0 = x;
      ref.y0 = y;
      ref.z0 = z;
      refHit.t = max_v;
      refHit.nx = refHit.ny = refHit.nz = zero_v;
      bounceColor = tracePacket(ref, refHit, 0.01f, bounces - 1, refIndex);
      reflectiveColor.r = spu_mul(bounceColor.r, reflectiveColor.r);
      reflectiveColor.g = spu_mul(bounceColor.g, reflectiveColor.g);
      reflectiveColor.b = spu_mul(bounceColor.b, reflectiveColor.b);
      reflectiveColor.r = spu_add(color.r, reflectiveColor.r);
      reflectiveColor.g = spu_add(color.g, reflectiveColor.g);
      reflectiveColor.b = spu_add(color.b, reflectiveColor.b);
      color.r = spu_sel(color.r, reflectiveColor.r, reflective);
      color.g = spu_sel(color.g, reflectiveColor.g, reflective);
      color.b = spu_sel(color.b, reflectiveColor.b, reflective);    
    }
#if (MAX_BOUNCES>1)
    // Do the same for refraction..
    RGBPacket transparentColor;
    transparentColor.r = transparentColor.g = transparentColor.b = zero_v;
    material_getTransparentColor(objects->materials, transparentColor, hit, x, y, z, changed);
    colorgtclip = spu_cmpgt(transparentColor.r, clipweight);
    colorgtclip = spu_or(colorgtclip, spu_cmpgt(transparentColor.g, clipweight));
    colorgtclip = spu_or(colorgtclip, spu_cmpgt(transparentColor.b, clipweight));
    colorgtclip = spu_gather(colorgtclip);
    vector unsigned int refractiver = spu_cmpgt(transparentColor.r, zero_v);
    vector unsigned int refractiveg = spu_cmpgt(transparentColor.g, zero_v);
    vector unsigned int refractiveb = spu_cmpgt(transparentColor.b, zero_v);
    vector unsigned int refractive;
    refractive = spu_or(refractiver, refractiveg);
    refractive = spu_or(refractive, refractiveb);
    refractive = spu_and(refractive, clipgthit);
    //refractive = spu_and(refractive, colorgtclip);
    recurse = (spu_extract(spu_gather(refractive), 0) > 0);
    if (recurse) {
      RayPacket refray;
      vector float negnx, negny, negnz;
      // TODO: I believe this method is not implemented yet.... parameters may change
      vector float RI;
      material_getRefractiveIndex(objects->materials,RI,  hit, x, y, z, changed);
      negnx = (vector float) spu_xor((vector unsigned int) hit.nx, sign_bit);
      negny = (vector float) spu_xor((vector unsigned int) hit.ny, sign_bit);
      negnz = (vector float) spu_xor((vector unsigned int) hit.nz, sign_bit);
      negnx = spu_sel(negnx, hit.nx, frontSide);
      negny = spu_sel(negny, hit.ny, frontSide);
      negnz = spu_sel(negnz, hit.nz, frontSide);
      vector float newRI = spu_sel(one_v, RI, frontSide);
      RI = spu_sel(RI, one_v, frontSide);
      vector unsigned int valid = transmit_vec3_v(refray.dx, refray.dy, refray.dz,
						  negnx, negny, negnz,
						  r.dx, r.dy, r.dz,
						  RI, newRI);
      refray.x0 = x;
      refray.y0 = y; 
      refray.z0 = z;
      HitPacket mHit;
      mHit.t = max_v; 
      mHit.nx = mHit.ny = mHit.nz = zero_v;
      RGBPacket bounceColor;
      valid = spu_and(valid, refractive);
      bounceColor = tracePacket(refray, mHit, 0.01f, bounces - 1, newRI);
      transparentColor.r = spu_mul(transparentColor.r, bounceColor.r);
      transparentColor.g = spu_mul(transparentColor.g, bounceColor.g);
      transparentColor.b = spu_mul(transparentColor.b, bounceColor.b);
      transparentColor.r = spu_add(transparentColor.r, color.r);
      transparentColor.g = spu_add(transparentColor.g, color.g);
      transparentColor.b = spu_add(transparentColor.b, color.b);
      color.r = spu_sel(color.r, transparentColor.r, valid);
      color.g = spu_sel(color.g, transparentColor.g, valid);
      color.b = spu_sel(color.b, transparentColor.b, valid);
    }
#endif
     }
#endif
  /*
  if (spu_extract(changed, 0)) {
    printf("tracePacket: hit updated: %f %f %f %f\n", spu_extract(hit.t, 0), spu_extract(hit.nx, 0), spu_extract(hit.ny, 0), spu_extract(hit.nz, 0));
  }
  */
  // For now, normal vizualization...
  /*
  color.r = spu_sel(zero_v, fabsf4(hit.nx), changed);
  color.g = spu_sel(zero_v, fabsf4(hit.ny), changed);
  color.b = spu_sel(zero_v, fabsf4(hit.nz), changed);
  */
  // ... and depth visualization
  /*
  color.r = spu_sel(zero_v, spu_mul(spu_sub(hit.t, one_v), half), changed);
  color.g = spu_sel(zero_v, spu_mul(spu_sub(hit.t, one_v), half), changed);
  color.b = spu_sel(zero_v, spu_mul(spu_sub(hit.t, one_v), half), changed);
  */
  return color;
}

